from itertools import chain, combinations

def get_constraints_for_labels(constraints,labels):
    constrains_for_label = dict()
    for label in labels:
        constrains_for_label[label] = []
    for constraint in constraints:
        constrains_for_label[constraint[1]].append(constraint[0] + constraint[2])
    return constrains_for_label


def powerset(iterable):
    "powerset([1,2,3]) --> () (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)"
    s = list(iterable)
    return chain.from_iterable(combinations(s, r) for r in range(len(s)+1))


#!/usr/bin/env python3

# sketch for an exponential algorithm determining whether problem is log* solvable or not
# missing to formally prove
# - iteratively building the set of equivalent labels by merging two sets is sound
# todo
#
# done
# - bugfix - removed separation of strongly connected components, fallback to try all subsets of labels
#   - assumption that log*n solvability of subset of labels implies log*n solvability for the whole strongly connected component is false (212 122 111 - by Aleksandr Tereshchenko)
# - print certificate for log*n solvability
# - alleviate assumption for reaching every other label
#   - by constructing a graph (hypergraph or normal with collapsed edges) and separate each strongly connected component
#       - "fine" as removal of a rule bridging s.c.c. cannot create problems because:
#           - either the "pointed" s.c.c. is log*n solvable, then such rule is useless.
#           - Otherwise, using a rule would force subtree to be ω(log n) solvable, so again useless.
import sys

VERBOSE = False

if __name__ == "__main__":
    from common import get_constraints_for_labels, powerset
else:
    from .common import get_constraints_for_labels, powerset



def is_log_star_solvable(constraints):
    labels = set("".join(constraints))
    log_star_solvable = False

    for reduced_labels in powerset(labels):
        reduced_constraints = [constraint for constraint in constraints if
                               (constraint[0] in reduced_labels and constraint[1] in reduced_labels and constraint[2] in reduced_labels)]
        if _is_log_star_solvable(reduced_constraints, list(reduced_labels)):
            log_star_solvable = True
            break
    return log_star_solvable



def _is_log_star_solvable(constraints, labels):
    root_labels = [[label] for label in labels]

    if VERBOSE:
        certificate = dict()

    constrains_for_labels = get_constraints_for_labels(constraints, labels)

    while True:
        root_labels_enlarged = False
        for e1 in root_labels:
            for e2 in root_labels:
                new_root_labels = []
                for label in labels:
                    for constraint in constrains_for_labels[label]:
                        if (constraint[0] in e1 and constraint[1] in e2) or \
                                (constraint[1] in e1 and constraint[0] in e2):
                            new_root_labels.append(label)
                            break
                if new_root_labels not in root_labels:
                    root_labels.append(new_root_labels)
                    if VERBOSE:
                        certificate[tuple(new_root_labels)] = (tuple(e1),tuple(e2))
                    root_labels_enlarged = True
                    break
            if root_labels_enlarged:
                break
        if not root_labels_enlarged:
            break

    if labels not in root_labels or not constraints:  # not constraints here for the edge-case of a single label
        return False
    else:
        if VERBOSE:
            print("Certificate:", certificate)
            print("digraph certificate {")
            for node, children in certificate.items():
                print(f'"{node}" -> "{children[0]}";')
                print(f'"{node}" -> "{children[1]}";')
            print("}")
        return True

if __name__ == "__main__":
    if len(sys.argv) == 2 and (sys.argv[1] == "-v" or sys.argv[1] == "--verbose"):
        VERBOSE = True
    constraints = input().split()
    if is_log_star_solvable(constraints):
        print("O(log*n)")
    else:
        print("ω(log*n)")



import sys

from .constant_decider import is_constant_solvable
from .log_decider import is_log_solvable
from .log_star_decider import is_log_star_solvable

if __name__ == "__main__":
    if len(sys.argv) == 2 and (sys.argv[1] == "-v" or sys.argv[1] == "--verbose"):
        VERBOSE = True
    constraints = input().split()
    if is_log_solvable(constraints):  # is not empty
        if is_log_star_solvable(constraints):
            if is_constant_solvable(constraints):
                print("O(1)")
            else:
                print("Θ(log*n)")
        else:
            print("Θ(log n)")
    else:
        print("Ω(n)")



import sys

if __name__ == "__main__":
    from common import powerset
    from log_star_decider import _is_log_star_solvable
else:
    from .common import powerset
    from .log_star_decider import _is_log_star_solvable


def is_constant_solvable(constraints):
    labels = set("".join(constraints))
    constant_solvable = False

    for reduced_labels in powerset(labels):
        reduced_constraints = [constraint for constraint in constraints if
                               (constraint[0] in reduced_labels and constraint[1] in reduced_labels and constraint[
                                   2] in reduced_labels)]
        has_aab = False
        for label in reduced_labels:
            for constraint in constraints:
                if constraint.startswith(label + label) or constraint.endswith(label + label):
                    has_aab = True

        if has_aab and _is_log_star_solvable(reduced_constraints, list(reduced_labels)):
            constant_solvable = True
            break
    return constant_solvable


if __name__ == "__main__":
    if len(sys.argv) == 2 and (sys.argv[1] == "-v" or sys.argv[1] == "--verbose"):
        VERBOSE = True
    constraints = input().split()
    if is_constant_solvable(constraints):
        print("O(1)")
    else:
        print("ω(1)")



#!/usr/bin/env python3

if __name__ == "__main__":
    from common import get_constraints_for_labels
else:
    from .common import get_constraints_for_labels

from math import gcd
from functools import reduce
import sys

VERBOSE = False

def listGcd(arr):
  return reduce(gcd, arr)

def countWalks(graph, src, dst, maxWalkLength):
    # Table to be filled up using DP.
    # The value count[i][j][e] will
    # store count of possible walks from
    # i to j with exactly k edges
    count = {i: {j: [0 for k in range(maxWalkLength + 1)] for j in graph} for i in graph}

    # Loop for number of edges from 0 to k
    for e in range(maxWalkLength + 1):
        for i in graph:  # for source
            for j in graph:  # for destination
                # initialize value
                count[i][j][e] = 0

                # from base cases
                if e == 0 and i == j:
                    count[i][j][e] = 1
                if e == 1 and (j in graph[i]):
                    count[i][j][e] = 1

                # go to adjacent only when the
                # number of edges is more than 1
                if e > 1:
                    for a in graph:
                        if a in graph[i]:  # adjacent of source i
                            count[i][j][e] += count[a][j][e - 1]
    return count[src][dst]

def isFlexible(graph, node):
    walkCounts = countWalks(graph, node, node, 2 * len(graph))
    bigL = []
    for idx, count in enumerate(walkCounts):
        if idx > 0 and count > 0:  # disregard a walk with 0 edges
            bigL.append(idx)
    return len(bigL) > 0 and listGcd(bigL) == 1


def inflexible_labels(constraints, labels):
    constraints_for_labels = get_constraints_for_labels(constraints,labels)
    graph = {label: [] for label in labels}
    for label in labels:
        for constraint in constraints_for_labels[label]:
            if constraint[0] not in graph[label]:
                graph[label].append(constraint[0])
            if constraint[1] not in graph[label]:
                graph[label].append(constraint[1])
    il = []
    for label in labels:
        if not isFlexible(graph, label):
            il.append(label)
    return il

def is_log_solvable(constraints):
    labels = list(set("".join(constraints)))

    while inflexible_labels(constraints, labels):
        il = inflexible_labels(constraints, labels)
        updated_constraints = []
        for constraint in constraints:
            if not (set(il) & set(constraint)): # keep only flexible labels
                updated_constraints.append(constraint)
        constraints = updated_constraints
        labels = list(set("".join(constraints)))
    if VERBOSE:
        print("Constraints:", constraints)
    return True if constraints else False

def log_decider(constraints):
    if is_log_solvable(constraints):  # is not empty
        print("O(log n)")
    else:
        print("Ω(n)")

if __name__ == "__main__":
    if len(sys.argv) == 2 and (sys.argv[1] == "-v" or sys.argv[1] == "--verbose"):
        VERBOSE = True
    constraints = input().split()
    log_decider(constraints)



import pathlib
from setuptools import setup

HERE = pathlib.Path(__file__).parent

README = (HERE / "README.md").read_text()

setup(
    name="rooted-tree-classifier",
    version="0.1.5",
    description="A command-line tool for automatically classifying LCL problems on rooted trees.",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/jendas1/rooted-tree-classifier",
    author="Jan Studený",
    author_email="jendas1@yahoo.com",
    license="Unlicense",
    classifiers=[
        "License :: OSI Approved :: The Unlicense (Unlicense)",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
    ],
    packages=["rooted_tree_classifier"],
    include_package_data=True,
    install_requires=["networkx"]
)
